@@ -173,8 +173,9 @@ getprogname (void)
 #include <fcntl.h>
 #include <unistd.h>
 
+#ifndef PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
 void
-p11_mutex_init (p11_mutex_t *mutex)
+p11_recursive_mutex_init (p11_mutex_t *mutex)
 {
 	pthread_mutexattr_t attr;
 	int ret;
@@ -185,6 +186,7 @@ p11_mutex_init (p11_mutex_t *mutex)
 	assert (ret == 0);
 	pthread_mutexattr_destroy (&attr);
 }
+#endif
 
 char *
 p11_dl_error (void)
@@ -527,7 +529,10 @@ strconcat (const char *first,
 	for (arg = first; arg; arg = va_arg (va, const char*)) {
 		size_t old_length = length;
 		length += strlen (arg);
-		return_val_if_fail (length >= old_length, NULL);
+		if (length < old_length) {
+			va_end (va);
+			return_val_if_reached (NULL);
+		}
 	}
 
 	va_end (va);
@@ -809,7 +814,7 @@ getauxval (unsigned long type)
 		extern int __libc_enable_secure;
 		secure = __libc_enable_secure;
 
-#elif defined(HAVE_ISSETUGID)
+#elif defined(HAVE_ISSETUGID) && defined(HAVE_ISSETUGID_OPENBSD)
 		secure = issetugid ();
 
 #elif defined(OS_UNIX)
@@ -914,6 +919,7 @@ fdwalk (int (* cb) (void *data, int fd),
 	struct rlimit rl;
 #endif
 
+#ifdef __linux__
 	dir = opendir ("/proc/self/fd");
 	if (dir != NULL) {
 		while ((de = readdir (dir)) != NULL) {
@@ -936,6 +942,7 @@ fdwalk (int (* cb) (void *data, int fd),
 		closedir (dir);
 		return res;
 	}
+#endif
 
 	/* No /proc, brute force */
 #ifdef HAVE_SYS_RESOURCE_H
@@ -957,3 +964,19 @@ fdwalk (int (* cb) (void *data, int fd),
 #endif /* HAVE_FDWALK */
 
 #endif /* OS_UNIX */
+
+int
+p11_ascii_tolower (int c)
+{
+	if (c >= 'A' && c <= 'Z')
+		return 'a' + (c - 'A');
+	return c;
+}
+
+int
+p11_ascii_toupper (int c)
+{
+	if (c >= 'a' && c <= 'z')
+		return 'A' + (c - 'a');
+	return c;
+}
